from lynara.interfaces.api_http import APIGatewayProxyEventV2Interface
from lynara.interfaces.api_rest import APIGatewayProxyEventV1Interface
from lynara.interfaces.lifespan import LifespanInterface

__all__ = [
    "APIGatewayProxyEventV1Interface",
    "APIGatewayProxyEventV2Interface",
    "LifespanInterface",
]



import pytest

from lynara import APIGatewayProxyEventV2Interface, Lynara


@pytest.mark.parametrize(
    "lifespan_mode",
    [
        "off",
        "auto",
    ],
)
async def test_django_app_lifespan_auto(lifespan_mode, lambda_events, django_app):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["requestContext"]["http"]["method"] = "GET"
    lambda_event["requestContext"]["http"]["path"] = "/django/"
    lambda_event["path"] = "/django/"
    lynara = Lynara(django_app, lifespan_mode=lifespan_mode)
    response = await lynara.run(lambda_event, None, APIGatewayProxyEventV2Interface)
    assert response["statusCode"] == 200
    assert response["body"] == "Hello, world!"


async def test_django_app_lifespan_on(lambda_events, django_app):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["requestContext"]["http"]["method"] = "GET"
    lambda_event["requestContext"]["http"]["path"] = "/django/"
    lambda_event["path"] = "/django/"
    lynara = Lynara(django_app, "on")
    with pytest.raises(ValueError) as cm:
        await lynara.run(lambda_event, None, APIGatewayProxyEventV2Interface)

    assert (
        str(cm.value) == "Django can only handle ASGI/HTTP connections, not lifespan."
    )



import json
from contextlib import asynccontextmanager
from copy import deepcopy
from pathlib import Path
from typing import Any
from unittest.mock import Mock

import pytest

from tests.apps.django_app import django_asgi_app
from tests.apps.fastapi_app import get_fast_api_app

BASE_PATH = Path(__file__).absolute().parent


@pytest.fixture()
def django_app():
    return django_asgi_app


class AsyncContextManager:
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def __call__(self, *args: Any, **kwds: Any) -> Any:
        pass


@pytest.fixture()
def mock_lifespan():
    return Mock()


@pytest.fixture()
def fastapi_app(mock_lifespan):
    @asynccontextmanager
    async def life(app):
        mock_lifespan(app, "startup")
        yield
        mock_lifespan(app, "shutdown")

    return get_fast_api_app(lifespan_func=life)


@pytest.fixture(scope="session")
def load_lambda_events():
    examples = {}
    for example_file in BASE_PATH.glob("event_examples/*"):
        if example_file.is_file():
            with open(example_file) as f:
                examples[example_file.stem] = json.load(f)
    return examples


@pytest.fixture()
def lambda_events(load_lambda_events):
    return deepcopy(load_lambda_events)



from lynara.interfaces.utils import get_server


def test_get_server():
    headers = {"host": "lynara:8080"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {"host": "lynara"}
    assert get_server(headers) == ("lynara", 80)

    headers = {"x-forwarded-port": "8080"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {}
    assert get_server(headers) == ("lynara", 80)

    headers = {"host": "lynara:8080", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {"host": "lynara", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 9090)

    headers = {"host": "lynara:8080", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {"host": "lynara", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 9090)

    headers = {"host": "lynara:8080", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {"host": "lynara", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 9090)

    headers = {"host": "lynara:8080", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 8080)

    headers = {"host": "lynara", "x-forwarded-port": "9090"}
    assert get_server(headers) == ("lynara", 9090)



from base64 import b64decode

import pytest

from lynara import APIGatewayProxyEventV2Interface
from tests.apps.fastapi_app import get_fast_api_app


async def test_init_no_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["body"] = None
    interface = APIGatewayProxyEventV2Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"",
        "more_body": False,
    }


async def test_init_non_base64_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["isBase64Encoded"] = False
    lambda_event["body"] = "Hello, world!"
    interface = APIGatewayProxyEventV2Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"Hello, world!",
        "more_body": False,
    }


async def test_init_bytes_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["isBase64Encoded"] = False
    lambda_event["body"] = b"Hello, world!"
    interface = APIGatewayProxyEventV2Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"Hello, world!",
        "more_body": False,
    }


async def test_scope(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    interface = APIGatewayProxyEventV2Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    scope = interface.scope

    assert scope["type"] == "http"
    assert scope["method"] == "POST"
    assert scope["path"] == "/resource"
    assert scope["query_string"] == (
        b"parameter1=value1&parameter1=value2&parameter2=value"
    )
    assert scope["headers"] == [
        (b"header1", b"value1"),
        (b"header2", b"value1,value2"),
    ]
    assert scope["client"] == ("192.168.0.1/32", 0)
    assert scope["server"] == ("lynara", 80)


async def test_match(lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    assert APIGatewayProxyEventV2Interface.match(lambda_event) is True
    lambda_event = lambda_events["api_gw_v1"]
    assert APIGatewayProxyEventV2Interface.match(lambda_event) is False


async def test_receive(lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    app = get_fast_api_app()
    interface = APIGatewayProxyEventV2Interface(app, lambda_event, context=None)

    message = await interface.receive()
    assert message["type"] == "http.request"
    assert message["body"] == b64decode(lambda_event["body"])
    assert not message["more_body"]


async def test_send_response_headers(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    interface = APIGatewayProxyEventV2Interface(fastapi_app, lambda_event, context=None)

    await interface.send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
            ],
        }
    )

    assert interface.lambda_response["statusCode"] == 200
    assert interface.lambda_response["headers"]["content-type"] == "application/json"


async def test_send_response_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    interface = APIGatewayProxyEventV2Interface(fastapi_app, lambda_event, context=None)

    await interface.send(
        {
            "type": "http.response.body",
            "body": b'{"message":"Hello, world!"}',
            "more_body": False,
        }
    )
    assert interface.lambda_response["body"] == '{"message":"Hello, world!"}'


async def test_send_response_body_more_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    interface = APIGatewayProxyEventV2Interface(fastapi_app, lambda_event, context=None)
    await interface.app_queue.get()

    await interface.send(
        {
            "type": "http.response.body",
            "body": b'{"message":"Hello, world!"}',
            "more_body": True,
        }
    )
    assert interface.lambda_response["body"] == '{"message":"Hello, world!"}'
    assert interface.app_queue.qsize() == 0


async def test_full_response(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["requestContext"]["http"]["method"] = "GET"
    interface = APIGatewayProxyEventV2Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )
    await interface.app_queue.get()

    await interface.send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
                (b"set-cookie", b"cookie1=value1"),
                (b"set-cookie", b"cookie2=value2"),
            ],
        }
    )
    await interface.send(
        {
            "type": "http.response.body",
            "body": b'{"Hello":"World"}',
            "more_body": False,
        }
    )

    response = interface.lambda_response

    assert response["statusCode"] == 200
    assert response["headers"]["content-type"] == "application/json"
    assert response["body"] == '{"Hello":"World"}'
    assert response["cookies"] == ["cookie1=value1", "cookie2=value2"]
    assert await interface.app_queue.get() == {
        "type": "http.disconnect",
    }


async def test_send_unknown_type(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    interface = APIGatewayProxyEventV2Interface(fastapi_app, lambda_event, context=None)

    with pytest.raises(ValueError):
        await interface.send(
            {
                "type": "unknown.type",
            }
        )



from collections.abc import Awaitable, Callable, MutableMapping
from enum import Enum
from typing import Any

LambdaEvent = dict[str, Any]
Scope = MutableMapping[str, Any]
Message = MutableMapping[str, Any]
Receive = Callable[[], Awaitable[Message]]
Send = Callable[[Message], Awaitable[None]]
ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]


class LifespanMode(str, Enum):
    AUTO = "auto"
    ON = "on"
    OFF = "off"



from lynara import APIGatewayProxyEventV1Interface


async def test_init_no_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["body"] = None
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"",
        "more_body": False,
    }


async def test_init_non_base64_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["isBase64Encoded"] = False
    lambda_event["body"] = "Hello, world!"
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"Hello, world!",
        "more_body": False,
    }


async def test_init_bytes_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["isBase64Encoded"] = False
    lambda_event["body"] = b"Hello, world!"
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, context=None, base_path="/path/to"
    )

    assert await interface.app_queue.get() == {
        "type": "http.request",
        "body": b"Hello, world!",
        "more_body": False,
    }


async def test_scope(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    lambda_event["httpMethod"] = "GET"
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, None, base_path="/path/to"
    )
    scope = interface.scope

    assert scope["type"] == "http"
    assert scope["method"] == "GET"
    assert scope["path"] == "/resource"
    assert scope["query_string"] == b"foo=bar"
    assert scope["headers"] == [
        [
            b"accept",
            b"text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*"
            b";q=0.8",
        ],
        [
            b"accept-encoding",
            b"gzip, deflate, sdch",
        ],
        [
            b"accept-language",
            b"en-US,en;q=0.8",
        ],
        [
            b"cache-control",
            b"max-age=0",
        ],
        [
            b"cloudfront-forwarded-proto",
            b"https",
        ],
        [
            b"cloudfront-is-desktop-viewer",
            b"true",
        ],
        [
            b"cloudfront-is-mobile-viewer",
            b"false",
        ],
        [
            b"cloudfront-is-smarttv-viewer",
            b"false",
        ],
        [
            b"cloudfront-is-tablet-viewer",
            b"false",
        ],
        [
            b"cloudfront-viewer-country",
            b"US",
        ],
        [
            b"host",
            b"0123456789.execute-api.us-east-1.amazonaws.com",
        ],
        [
            b"upgrade-insecure-requests",
            b"1",
        ],
        [
            b"user-agent",
            b"Custom User Agent String",
        ],
        [
            b"via",
            b"1.1 08f323deadbeefa7af34d5feb414ce27.cloudfront.net (CloudFront)",
        ],
        [
            b"x-amz-cf-id",
            b"cDehVQoZnx43VYQb9j2-nvCh-9z396Uhbp027Y2JvkCPNLmGJHqlaA==",
        ],
        [
            b"x-forwarded-for",
            b"127.0.0.1, 127.0.0.2",
        ],
        [
            b"x-forwarded-port",
            b"443",
        ],
        [
            b"x-forwarded-proto",
            b"https",
        ],
    ]


async def test_match(lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    assert APIGatewayProxyEventV1Interface.match(lambda_event) is True
    lambda_event = lambda_events["api_gw_v2"]
    assert APIGatewayProxyEventV1Interface.match(lambda_event) is False


async def test_receive(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, None, base_path="/path/to"
    )
    message = await interface.receive()
    assert message["type"] == "http.request"
    assert message["body"] == b'{"test":"body"}'
    assert not message["more_body"]


async def test_send_response_headers(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, None, base_path="/path/to"
    )

    await interface.send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
                (b"set-cookie", b"cookie1=value1"),
                (b"set-cookie", b"cookie2=value2"),
            ],
        }
    )

    assert interface.lambda_response["statusCode"] == 200
    assert interface.lambda_response["headers"]["content-type"] == "application/json"
    assert "set-cookie" in interface.lambda_response["multiValueHeaders"]
    assert interface.lambda_response["multiValueHeaders"]["set-cookie"] == [
        "cookie1=value1",
        "cookie2=value2",
    ]


async def test_send_response_body(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, None, base_path="/path/to"
    )

    await interface.send(
        {
            "type": "http.response.body",
            "body": b'{"Hello": "World"}',
            "more_body": False,
        }
    )

    assert interface.lambda_response["body"] == '{"Hello": "World"}'


async def test_full_response(fastapi_app, lambda_events):
    lambda_event = lambda_events["api_gw_v1"]
    lambda_event["httpMethod"] = "GET"
    interface = APIGatewayProxyEventV1Interface(
        fastapi_app, lambda_event, None, base_path="/path/to"
    )

    await interface.send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [
                (b"content-type", b"application/json"),
                (b"set-cookie", b"cookie1=value1"),
                (b"set-cookie", b"cookie2=value2"),
            ],
        }
    )
    await interface.send(
        {
            "type": "http.response.body",
            "body": b'{"Hello":"World"}',
            "more_body": False,
        }
    )

    response = interface.lambda_response

    assert response["statusCode"] == 200
    assert response["headers"]["content-type"] == "application/json"
    assert response["multiValueHeaders"]["set-cookie"] == [
        "cookie1=value1",
        "cookie2=value2",
    ]
    assert response["body"] == '{"Hello":"World"}'



from abc import ABC, abstractmethod
from asyncio import Queue
from typing import Any

from lynara.types import ASGIApp, LambdaEvent, Message, Scope


class HTTPInterface(ABC):
    event: LambdaEvent
    context: Any
    is_response_completed: bool
    lambda_response: dict[str, Any]

    @classmethod
    @abstractmethod
    def match(cls, event: LambdaEvent) -> bool:
        raise NotImplementedError

    def __init__(
        self, app: ASGIApp, event: LambdaEvent, context, base_path: str | None = None
    ) -> None:
        self.app = app
        self.event = event
        self.context = context
        self.is_response_completed = False
        self._method: str | None = None
        self.lambda_response = {}
        self.app_queue: Queue[Message] = Queue()
        self.base_path = base_path

    async def __call__(self) -> Any:
        await self.app(self.scope, self.receive, self.send)
        return self.lambda_response

    @property
    @abstractmethod
    def scope(self) -> Scope:
        raise NotImplementedError

    @abstractmethod
    async def receive(self) -> Message:
        raise NotImplementedError

    @abstractmethod
    async def send(self, message: Message) -> None:
        raise NotImplementedError



from lynara.interfaces import (
    APIGatewayProxyEventV1Interface,
    APIGatewayProxyEventV2Interface,
    LifespanInterface,
)
from lynara.runner import Lynara

__all__ = [
    "Lynara",
    "APIGatewayProxyEventV2Interface",
    "APIGatewayProxyEventV1Interface",
    "LifespanInterface",
]



from base64 import b64decode
from urllib.parse import urlencode

from lynara.interfaces.base import HTTPInterface
from lynara.interfaces.utils import get_server, strip_api_gateway_path
from lynara.types import ASGIApp, LambdaEvent, Message, Scope


class APIGatewayProxyEventV1Interface(HTTPInterface):
    """
    `sam local generate-event apigateway aws-proxy`
    """

    def __init__(
        self, app: ASGIApp, event: LambdaEvent, context, base_path: str | None = None
    ) -> None:
        super().__init__(app=app, event=event, context=context, base_path=base_path)
        self.lambda_response = {
            "isBase64Encoded": False,
            "statusCode": 200,
            "headers": {},
            "multiValueHeaders": {},
            "body": "",
        }
        body = self.event.get("body", b"")
        if not body:
            body = b""
        if self.event.get("isBase64Encoded"):
            body = b64decode(body)
        elif not isinstance(body, bytes):
            body = body.encode()
        self.app_queue.put_nowait(
            {
                "type": "http.request",
                "body": body,
                "more_body": False,
            }
        )

    @classmethod
    def match(cls, event: LambdaEvent) -> bool:
        return "resource" in event and "requestContext" in event

    def _handle_multi_value_headers_in_scope(self) -> dict[str, str]:
        headers = self.event.get("headers", {}) or {}
        headers = {k.lower(): v for k, v in headers.items()}
        if self.event.get("multiValueHeaders"):
            headers.update(
                {
                    k.lower(): ", ".join(v) if isinstance(v, list) else ""
                    for k, v in self.event.get("multiValueHeaders", {}).items()
                }
            )

        return headers

    def _encode_query_string(self) -> bytes:
        params = self.event.get("multiValueQueryStringParameters", {})
        if not params:
            params = self.event.get("queryStringParameters", {})
        if not params:
            return b""

        return urlencode(params, doseq=True).encode()

    @property
    def scope(self) -> Scope:
        headers = self._handle_multi_value_headers_in_scope()
        request_context = self.event["requestContext"]
        self._method = self.event["httpMethod"]
        return {
            "type": "http",
            "method": self._method,
            "http_version": "1.1",
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "path": strip_api_gateway_path(
                self.event["path"],
                base_path=self.base_path,
            ),
            "raw_path": None,
            "root_path": "",
            "scheme": headers.get("x-forwarded-proto", "https"),
            "query_string": self._encode_query_string(),
            "server": get_server(headers=headers),
            "client": (request_context.get("identity", {}).get("sourceIp"), 0),
            "asgi": {
                "version": "3.0",
                "spec_version": "2.3",
            },
        }

    async def receive(self) -> Message:
        return await self.app_queue.get()

    def _handle_multi_value_headers(
        self, response_headers: list[tuple[bytes, bytes]]
    ) -> tuple[dict[str, str], dict[str, list[str]]]:
        headers: dict[str, str] = {}
        multi_value_headers: dict[str, list[str]] = {}

        for key, value in response_headers:
            decoded_key = key.decode().lower()
            decoded_value = value.decode()

            if decoded_key in multi_value_headers:
                multi_value_headers[decoded_key].append(decoded_value)
            elif decoded_key in headers:
                multi_value_headers[decoded_key] = [
                    headers.pop(decoded_key),
                    decoded_value,
                ]
            else:
                headers[decoded_key] = decoded_value

        return headers, multi_value_headers

    async def send(self, message: Message) -> None:
        if message["type"] == "http.response.start":
            self.lambda_response["statusCode"] = message["status"]
            headers, multi_value_headers = self._handle_multi_value_headers(
                message.get("headers", [])
            )
            self.lambda_response["headers"] = headers
            self.lambda_response["multiValueHeaders"] = multi_value_headers

        elif message["type"] == "http.response.body":
            self.lambda_response["body"] += message.get("body", b"").decode()
            more_body = message.get("more_body", False)

            if not more_body:
                await self.app_queue.put({"type": "http.disconnect"})

        else:
            raise ValueError(f"Unknown message type: {message['type']}")



from unittest.mock import call

import pytest

from lynara import APIGatewayProxyEventV2Interface, Lynara
from lynara.types import LifespanMode


@pytest.mark.parametrize(
    "lifespan_mode",
    [
        "on",
        "auto",
    ],
)
async def test_fastapi_app_lifespan(
    lifespan_mode, lambda_events, mock_lifespan, fastapi_app
):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["requestContext"]["http"]["method"] = "GET"
    lambda_event["requestContext"]["http"]["path"] = "/fastapi/"
    lambda_event["path"] = "/fastapi/"
    lynara = Lynara(fastapi_app, LifespanMode(lifespan_mode))
    response = await lynara.run(lambda_event, None, APIGatewayProxyEventV2Interface)

    assert response["statusCode"] == 200
    assert response["body"] == '"Hello, world!"'
    mock_lifespan.assert_has_calls(
        [call(fastapi_app, "startup"), call(fastapi_app, "shutdown")]
    )
    assert response["cookies"] == ["cookie=test; Path=/; SameSite=lax"]
    assert response["headers"] == {
        "content-length": "15",
        "content-type": "application/json",
        "x-custom-header": "test",
    }


async def test_fastapi_app_lifespan_off(lambda_events, mock_lifespan, fastapi_app):
    lambda_event = lambda_events["api_gw_v2"]
    lambda_event["requestContext"]["http"]["method"] = "GET"
    lambda_event["requestContext"]["http"]["path"] = "/fastapi/"
    lambda_event["path"] = "/fastapi/"
    lynara = Lynara(fastapi_app, LifespanMode.OFF)
    response = await lynara.run(lambda_event, None, APIGatewayProxyEventV2Interface)

    assert response["statusCode"] == 200
    assert response["body"] == '"Hello, world!"'
    mock_lifespan.assert_not_called()



from typing import Any
from urllib.parse import unquote


def get_server(headers: dict[str, Any]) -> tuple[str, int]:
    server_name = headers.get("host", "lynara")
    parts = server_name.split(":")
    if len(parts) == 2:
        server_name = parts[0]
        server_port = parts[1]
    else:
        server_port = headers.get("x-forwarded-port", 80)
    return (server_name, int(server_port))


def strip_api_gateway_path(path: str, *, base_path: str | None = None) -> str:
    if not path:
        return "/"

    normalized_base_path = (
        f"/{base_path.lstrip('/')}" if base_path and base_path != "/" else ""
    )

    if path.startswith(normalized_base_path):
        path = path[len(normalized_base_path) :]

    return unquote(path)



from fastapi import FastAPI, Response


def get_fast_api_app(lifespan_func=None):
    fastapi_asgi_app = FastAPI(lifespan=lifespan_func)

    @fastapi_asgi_app.get("/fastapi/")
    @fastapi_asgi_app.get("/resource")
    async def read_root(response: Response):
        response.set_cookie(key="cookie", value="test")
        response.headers["X-Custom-Header"] = "test"
        return "Hello, world!"

    return fastapi_asgi_app


if __name__ == "__main__":
    import asyncio

    from lynara import APIGatewayProxyEventV1Interface, Lynara

    lynara = Lynara(app=get_fast_api_app())
    response = asyncio.run(
        lynara.run(
            event={
                "version": "1.0",
                "resource": "/dev/fastapi/",
                "path": "/dev/fastapi/",
                "httpMethod": "GET",
                "headers": {"header1": "value1", "header2": "value2"},
                "multiValueHeaders": {
                    "header3": ["value3"],
                    "header4": ["value4", "value5"],
                },
                "queryStringParameters": {
                    "parameter1": "value1",
                    "parameter2": "value",
                },
                "multiValueQueryStringParameters": {
                    "parameter1": ["value1", "value2"],
                    "parameter2": ["value"],
                },
                "requestContext": {
                    "accountId": "123456789012",
                    "apiId": "id",
                    "authorizer": {"claims": None, "scopes": None},
                    "domainName": "id.execute-api.us-east-1.amazonaws.com",
                    "domainPrefix": "id",
                    "extendedRequestId": "request-id",
                    "httpMethod": "GET",
                    "identity": {
                        "accessKey": None,
                        "accountId": None,
                        "caller": None,
                        "cognitoAuthenticationProvider": None,
                        "cognitoAuthenticationType": None,
                        "cognitoIdentityId": None,
                        "cognitoIdentityPoolId": None,
                        "principalOrgId": None,
                        "sourceIp": "192.0.2.1",
                        "user": None,
                        "userAgent": "user-agent",
                        "userArn": None,
                        "clientCert": {
                            "clientCertPem": "CERT_CONTENT",
                            "subjectDN": "www.example.com",
                            "issuerDN": "Example issuer",
                            "serialNumber": (
                                "a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1:a1"
                            ),
                            "validity": {
                                "notBefore": "May 28 12:30:02 2019 GMT",
                                "notAfter": "Aug  5 09:36:04 2021 GMT",
                            },
                        },
                    },
                    "path": "/dev/fastapi/",
                    "protocol": "HTTP/1.1",
                    "requestId": "id=",
                    "requestTime": "04/Mar/2020:19:15:17 +0000",
                    "requestTimeEpoch": 1583349317135,
                    "resourceId": None,
                    "resourcePath": "/dev/fastapi/",
                    "stage": "dev",
                },
                "pathParameters": None,
                "stageVariables": None,
                "body": "Hello from Lambda!",
                "isBase64Encoded": False,
            },
            context=None,
            interface_class=APIGatewayProxyEventV1Interface,
            base_path="/dev",
        )
    )
    print(response)  # noqa: T201



import asyncio
import logging

from lynara.types import ASGIApp, LifespanMode, Message

LOGGER = logging.getLogger(__name__)


class LifespanInterface:
    def __init__(self, app: ASGIApp, lifespan_mode: LifespanMode) -> None:
        self.app = app
        self.lifespan_mode = lifespan_mode
        self.validate_mode()
        self.scope = {
            "type": "lifespan",
            "asgi": {"spec_version": "1.0", "version": "3.0"},
            "state": {},
        }
        self._queue: asyncio.Queue[Message] = asyncio.Queue()
        self._startup_event = asyncio.Event()
        self._shutdown_event = asyncio.Event()
        self.error_occured = False
        self.startup_failed = False
        self.shutdown_failed = False

    def validate_mode(self) -> None:
        if self.lifespan_mode == LifespanMode.OFF:
            raise Exception("Lifespan mode is off, but lifespan interface is used")

    async def __aenter__(self) -> "LifespanInterface":
        await self.startup()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.shutdown()

    async def startup(self) -> None:
        loop = asyncio.get_event_loop()
        main_lifespan_task = loop.create_task(self.main())
        await self._queue.put({"type": "lifespan.startup"})
        await self._startup_event.wait()

        if self.startup_failed or (
            self.error_occured and self.lifespan_mode == LifespanMode.ON
        ):
            LOGGER.error("Application startup failed. Exiting.")
            possible_exception = main_lifespan_task.exception()
            if isinstance(possible_exception, Exception):
                raise possible_exception

    async def shutdown(self):
        await self._queue.put({"type": "lifespan.shutdown"})
        LOGGER.debug("Waiting for shutdown")
        await self._shutdown_event.wait()
        LOGGER.debug("Shutdown complete")

    async def main(self):
        try:
            await self.app(scope=self.scope, receive=self.receive, send=self.send)
        except BaseException as exc:
            self.error_occured = True
            if self.lifespan_mode == LifespanMode.AUTO:
                LOGGER.info("Lifespan protocol appears to be unsupported, skipping.")
            else:
                LOGGER.exception(
                    "Error in lifespan interface, does the app support lifespan?"
                )
            raise exc
        finally:
            self._startup_event.set()
            self._shutdown_event.set()

    async def receive(self) -> Message:
        return await self._queue.get()

    async def send(self, message: Message) -> None:
        message_type = message["type"]
        if message_type == "lifespan.startup.complete":
            self._startup_event.set()
            return
        elif message_type == "lifespan.startup.failed":
            self.startup_failed = True
            self._startup_event.set()
            raise Exception(f'Startup failed {message.get("message", "")}')
        elif message_type == "lifespan.shutdown.complete":
            self._shutdown_event.set()
            return
        elif message_type == "lifespan.shutdown.failed":
            self.shutdown_failed = True
            self._shutdown_event.set()
            raise Exception(f'Shutdown failed {message.get("message", "")}')
        raise Exception(f"Invalid startup message type {message_type}")



from unittest.mock import AsyncMock, Mock, patch

from lynara.interfaces import APIGatewayProxyEventV2Interface, LifespanInterface
from lynara.runner import Lynara
from lynara.types import LifespanMode


async def test_runner(lambda_events):
    lambda_event = lambda_events["api_gw_v2"]
    mock_app = AsyncMock()
    mock_interface = AsyncMock()
    mock_interface_class = Mock(spec=APIGatewayProxyEventV2Interface)
    mock_interface_class.return_value = mock_interface
    mock_lifespan = AsyncMock()

    lynara = Lynara(mock_app)

    with patch(
        "lynara.runner.LifespanInterface", spec=LifespanInterface
    ) as mock_lifespan_class:
        mock_lifespan_class.return_value = mock_lifespan
        await lynara.run(lambda_event, None, mock_interface_class)

    mock_interface_class.assert_called_once_with(
        app=mock_app, event=lambda_event, context=None, base_path=None
    )
    mock_interface.assert_called_once()
    mock_lifespan_class.assert_called_once_with(
        app=mock_app,
        lifespan_mode=LifespanMode.AUTO,
    )



import logging
from contextlib import AsyncExitStack
from time import time

from lynara import LifespanInterface
from lynara.interfaces.base import HTTPInterface
from lynara.types import LifespanMode

LOGGER = logging.getLogger(__name__)


class Lynara:
    def __init__(self, app, lifespan_mode: LifespanMode = LifespanMode.AUTO):
        self.app = app
        self.lifespan_mode = lifespan_mode

    async def run(
        self,
        event,
        context,
        interface_class: type[HTTPInterface],
        base_path: str | None = None,
    ):
        start_time = time()
        interface = interface_class(
            app=self.app, event=event, context=context, base_path=base_path
        )
        async with AsyncExitStack() as stack:
            if self.lifespan_mode in (LifespanMode.ON, LifespanMode.AUTO):
                await stack.enter_async_context(
                    LifespanInterface(app=self.app, lifespan_mode=self.lifespan_mode)
                )
            interface_start_time = time()
            lambda_response = await interface()
        LOGGER.info(
            "Lynara execution time: %.5f s, out of which interface time: %.5f s",
            (time() - start_time),
            (time() - interface_start_time),
        )
        return lambda_response



from base64 import b64decode

from lynara.interfaces.base import HTTPInterface
from lynara.interfaces.utils import get_server, strip_api_gateway_path
from lynara.types import ASGIApp, LambdaEvent, Message, Scope


class APIGatewayProxyEventV2Interface(HTTPInterface):
    def __init__(
        self, app: ASGIApp, event: LambdaEvent, context, base_path: str | None = None
    ) -> None:
        super().__init__(app=app, event=event, context=context, base_path=base_path)
        self.lambda_response = {
            "cookies": [],
            "isBase64Encoded": False,
            "statusCode": 200,
            "body": "",
            "headers": {},
        }

        body = self.event.get("body", b"")
        if not body:
            body = b""
        if self.event.get("isBase64Encoded"):
            body = b64decode(body)
        elif not isinstance(body, bytes):
            body = body.encode()
        self.app_queue.put_nowait(
            {
                "type": "http.request",
                "body": body,
                "more_body": False,
            }
        )

    @classmethod
    def match(cls, event: LambdaEvent) -> bool:
        return event.get("version") == "2.0" and "requestContext" in event

    @property
    def scope(self) -> Scope:
        headers = {k.lower(): v for k, v in self.event.get("headers", {}).items()}
        request_context = self.event["requestContext"]
        path = request_context["http"]["path"]
        self._method = request_context["http"]["method"]

        return {
            "type": "http",
            "asgi": {
                "version": "3.0",
                "spec_version": "2.3",
            },
            "http_version": "1.1",
            "method": self._method,
            "scheme": headers.get("x-forwarded-proto", "https"),
            "path": strip_api_gateway_path(path, base_path=self.base_path),
            "raw_path": None,
            "query_string": self.event.get("rawQueryString", "").encode(),
            "root_path": "",
            "headers": [(k.encode(), v.encode()) for k, v in headers.items()],
            "client": (request_context["http"]["sourceIp"], 0),
            "server": get_server(headers=headers),
        }

    async def receive(self) -> Message:
        return await self.app_queue.get()

    async def send(self, message: Message) -> None:
        if message["type"] == "http.response.start":
            self.lambda_response["statusCode"] = message["status"]
            for key, value in message.get("headers", []):
                if key.decode().lower() == "set-cookie":
                    self.lambda_response["cookies"].append(value.decode())
                else:
                    self.lambda_response["headers"][key.decode()] = value.decode()

        elif message["type"] == "http.response.body":
            self.lambda_response["body"] += message.get("body", b"").decode()
            more_body = message.get("more_body", False)

            if not more_body:
                await self.app_queue.put({"type": "http.disconnect"})

        else:
            raise ValueError(f"Unknown message type: {message['type']}")



from django.conf import settings
from django.core.asgi import get_asgi_application
from django.http import HttpResponse
from django.urls import path

settings.configure(
    DEBUG=True,
    SECRET_KEY="thisisthesecretkey",
    ALLOWED_HOSTS=["*"],
    ROOT_URLCONF=__name__,
)


def index(request):
    return HttpResponse("Hello, world!")


urlpatterns = [
    path("django/", index),
]

django_asgi_app = get_asgi_application()


if __name__ == "__main__":
    import asyncio

    from lynara import APIGatewayProxyEventV2Interface, Lynara

    lynara = Lynara(app=django_asgi_app)
    response = asyncio.run(
        lynara.run(
            event={
                "version": "2.0",
                "routeKey": "$default",
                "rawPath": "/django/",
                "rawQueryString": (
                    "parameter1=value1&" "parameter1=value2&" "parameter2=value"
                ),
                "cookies": ["cookie1", "cookie2"],
                "headers": {"Header1": "value1", "Header2": "value1,value2"},
                "queryStringParameters": {
                    "parameter1": "value1,value2",
                    "parameter2": "value",
                },
                "requestContext": {
                    "domainName": "id.execute-api.us-east-1.amazonaws.com",
                    "domainPrefix": "id",
                    "http": {
                        "method": "GET",
                        "path": "/django/",
                        "protocol": "HTTP/1.1",
                        "sourceIp": "192.168.0.1/32",
                        "userAgent": "agent",
                    },
                    "requestId": "id",
                    "routeKey": "$default",
                    "stage": "$default",
                    "time": "12/Mar/2020:19:03:58 +0000",
                    "timeEpoch": 1583348638390,
                },
                "body": "eyJ0ZXN0IjoiYm9keSJ9",
                "pathParameters": {"parameter1": "value1"},
                "isBase64Encoded": True,
                "stageVariables": {
                    "stageVariable1": "value1",
                    "stageVariable2": "value2",
                },
            },
            context=None,
            interface_class=APIGatewayProxyEventV2Interface,
        )
    )
